from .shader import Filter2D


glsl = """
uniform sampler2D bgl_RenderedTexture;

in vec4 bgl_TexCoord;
uniform float power;

out vec4 fragColor;

void main()
{
    float pow = clamp(1.0 - power, 0.0, 1.0);
	vec4 px =  texture(bgl_RenderedTexture, bgl_TexCoord.xy);
	float grey = 0.21 * px.r + 0.71 * px.g + 0.07 * px.b;
	fragColor = vec4(
        px.r * pow + grey * (1.0 - pow),
        px.g * pow + grey * (1.0 - pow),
        px.b * pow + grey * (1.0 - pow),
        1.0
    );
}
"""


class Grayscale(Filter2D):

    def __init__(self, power: float = 1.0, idx: int = None) -> None:
        self.settings = {'power': float(power)}
        super().__init__(glsl, idx, {'power': self.settings, 'color': self.settings})

    @property
    def power(self):
        return self.settings['power']

    @power.setter
    def power(self, val):
        self.settings['power'] = val



'''The Uplogic Animation Module provides easy to use API for playing and
controlling animations.
'''

from .actionsystem import ULActionSystem  # noqa
from .action import ACTION_FINISHED  # noqa
from .action import ACTION_STARTED  # noqa
from .action import ULAction, Action  # noqa
from .sequence import ULSequence, Sequence  # noqa



from bge.types import KX_GameObject as GameObject
from bge import logic
import bpy
from mathutils import Vector, Matrix, Color


def make_unique_light(light: GameObject) -> GameObject:
    '''Make a copy of this light to allow for unique modifications.

    :param `light`: The `KX_LightObject` of which to copy the data.

    :returns: The modified `KX_LightObject`. This return value is for usability reasons and
    is the same as the original argument.
    '''
    lamp = light.blenderObject
    lamp.data = lamp.data.copy()
    return light


class ULLight():
    """Wrapper for `KX_LightObject`. Can either convert a given scene light or create a new one.

    :param `name`: The name of the new light (only relevant if `light` argument is None).
    :param `type`: The type of the new light of [`"POINT"`, `"SUN"`, `"SPOT"`, `"AREA"`] (only relevant if `light` argument is None).
    :param `light`: The scene light to be converted. Leave at `None` to create new.
    """

    _deprecated = True

    def __init__(
        self,
        name: str = '',
        type: str = 'POINT',
        light: GameObject = None
    ) -> None:
        if self._deprecated:
            print('[UPLOGIC] ULLight class will be renamed to "Light" in future releases!')
        if light:
            self.light = make_unique_light(light)
            """The wrapped `KX_LightObject`."""
            return
        game_scene = logic.getCurrentScene()
        scene = bpy.data.scenes[game_scene.name]
        light = bpy.data.lights.new(name, type)
        light = bpy.data.objects.new(name, light)
        scene.collection.objects.link(light)
        self.light = game_scene.convertBlenderObject(light)
        """The wrapped `KX_LightObject`."""
        self.energy = 10

    @property
    def energy(self) -> float:
        return self.light.blenderObject.data.energy

    @energy.setter
    def energy(self, val: float):
        self.light.blenderObject.data.energy = val

    @property
    def color(self) -> Color:
        return self.light.blenderObject.data.color

    @color.setter
    def color(self, val: Color):
        self.light.blenderObject.data.color = val

    @property
    def use_shadow(self) -> bool:
        return self.light.blenderObject.data.use_shadow

    @use_shadow.setter
    def use_shadow(self, val: bool):
        self.light.blenderObject.data.use_shadow = val

    @property
    def use_custom_distance(self) -> bool:
        return self.light.blenderObject.data.use_custom_distance

    @use_custom_distance.setter
    def use_custom_distance(self, val: bool):
        self.light.blenderObject.data.use_custom_distance = val

    @property
    def distance(self) -> float:
        return self.light.blenderObject.data.cutoff_distance

    @distance.setter
    def distance(self, val: float):
        self.light.blenderObject.data.cutoff_distance = val

    @property
    def angle(self)-> float:
        return self.light.blenderObject.data.angle

    @angle.setter
    def angle(self, val: float):
        self.light.blenderObject.data.angle = val

    @property
    def spot_size(self) -> float:
        return self.light.blenderObject.data.spot_size

    @spot_size.setter
    def spot_size(self, val: float):
        self.light.blenderObject.data.spot_size = val

    @property
    def spot_blend(self) -> float:
        return self.light.blenderObject.data.spot_blend

    @spot_blend.setter
    def spot_blend(self, val: float):
        self.light.blenderObject.data.spot_blend = val

    @property
    def parent(self) -> GameObject:
        return self.light.parent

    @parent.setter
    def parent(self, val: GameObject):
        self.light.setParent(val)

    @property
    def worldPosition(self) -> Vector:
        return self.light.worldPosition

    @worldPosition.setter
    def worldPosition(self, val: Vector):
        self.light.worldPosition = val

    @property
    def localPosition(self) -> Vector:
        return self.light.localPosition

    @localPosition.setter
    def localPosition(self, val: Vector):
        self.light.localPosition = val

    @property
    def worldOrientation(self) -> Matrix:
        return self.light.worldOrientation

    @worldOrientation.setter
    def worldOrientation(self, val: Matrix):
        self.light.worldOrientation = val

    @property
    def localOrientation(self) -> Matrix:
        return self.light.localOrientation

    @localOrientation.setter
    def localOrientation(self, val: Matrix):
        self.light.localOrientation = val

    @property
    def worldScale(self) -> Vector:
        return self.light.worldScale

    @worldScale.setter
    def worldScale(self, val: Vector):
        self.light.worldScale = val

    @property
    def localScale(self) -> Vector:
        return self.light.localScale

    @localScale.setter
    def localScale(self, val: Vector):
        self.light.localScale = val

    @property
    def worldLinearVelocity(self) -> Vector:
        return self.light.worldLinearVelocity

    @worldLinearVelocity.setter
    def worldLinearVelocity(self, val: Vector):
        self.light.worldLinearVelocity = val

    @property
    def localLinearVelocity(self) -> Vector:
        return self.light.localLinearVelocity

    @localLinearVelocity.setter
    def localLinearVelocity(self, val: Vector):
        self.light.localLinearVelocity = val

    @property
    def worldAngularVelocity(self) -> Vector:
        return self.light.worldAngularVelocity

    @worldAngularVelocity.setter
    def worldAngularVelocity(self, val: Vector):
        self.light.worldAngularVelocity = val

    @property
    def localAngularVelocity(self) -> Vector:
        return self.light.localAngularVelocity

    @localAngularVelocity.setter
    def localAngularVelocity(self, val: Vector):
        self.light.localAngularVelocity = val

    @property
    def worldTransform(self) -> Matrix:
        return self.light.worldTransform

    @worldTransform.setter
    def worldTransform(self, val: Matrix):
        self.light.worldTransform = val


class Light(ULLight):
    _deprecated = False



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.ui import Image
from uplogic.utils import not_met


class ULCreateUIImage(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.parent = None
        self.rel_pos = None
        self.pos = None
        self.rel_size = None
        self.size = None
        self.angle = None
        self.texture = None
        self._widget = None
        self.layout_type = 'RelativeLayout'
        self.boxlayout_type = 'vertical'
        self.halign_type = 'left'
        self.valign_type = 'bottom'
        self._done = False
        self.OUT = ULOutSocket(self, self._get_done)
        self.WIDGET = ULOutSocket(self, self._get_widget)

    def _get_done(self):
        return self._done

    def _get_widget(self):
        return self._widget

    def evaluate(self):
        self._done = False
        condition = self.get_input(self.condition)
        self._set_ready()
        if not_met(condition):
            return
        ipt = self.get_input
        parent = ipt(self.parent)
        rel_pos = ipt(self.rel_pos)
        pos = ipt(self.pos)
        rel_size = ipt(self.rel_size)
        size = ipt(self.size)
        texture = ipt(self.texture)

        self._widget = Image(
            pos=pos,
            size=size,
            texture=texture,
            relative={'pos': rel_pos, 'size': rel_size},
            halign=self.halign_type,
            valign=self.valign_type
        )
        if parent:
            parent.add_widget(self._widget)
        self._done = True
        



from .widget import Widget
import blf
from bge import render
from bpy.types import VectorFont


class Label(Widget):
    """Widget for displaying text"""

    def __init__(
        self,
        pos=[0, 0],
        bg_color=[0, 0, 0, 0],
        relative={},
        text='',
        font='',
        font_color=[1, 1, 1, 1],
        font_size=12,
        line_height=1.5,
        shadow=False,
        shadow_offset=[2, -2],
        shadow_color=[0, 0, 0, .6],
        halign='left',
        valign='bottom',
        wrap=False
    ):
        """
        :param parent: the widget's parent
        :param name: the name of the widget
        :param text: the text to display (this can be changed later via the text property)
        :param font: the font to use
        :param pt_size: the point size of the text to draw (defaults to 30 if None)
        :param bg_color: the color to use when rendering the font
        :param pos: a tuple containing the x and y position
        :param sub_theme: name of a sub_theme defined in the theme file (similar to CSS classes)
        :param options: various other options

        """
        self._parent = None
        self._children = None
        self._font_color = font_color
        self.text = text
        self.line_height = line_height
        self.shadow = shadow
        self.shadow_offset = shadow_offset
        self.shadow_color = shadow_color
        self.bg_color = bg_color
        self.font_size = font_size
        self.font_color = font_color
        self.font = font
        self.wrap = wrap
        Widget.__init__(self, pos, (0, 0), bg_color, relative)
        self.text_halign = halign
        self.text_valign = valign

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, val):
        self._text = str(val)

    @property
    def font(self):
        return self._font

    @font.setter
    def font(self, val):
        if isinstance(val, VectorFont):
            val = val.filepath.replace('\\', '/')
        self._font = blf.load(val) if val else 0

    @property
    def font_color(self):
        return self._font_color

    @font_color.setter
    def font_color(self, val):
        val = list(val)
        self._font_color = val

    @property
    def opacity(self):
        return self.bg_color[3]

    @opacity.setter
    def opacity(self, val):
        self.bg_color[3] = val
        self.font_opacity = val

    @property
    def font_opacity(self):
        return self._font_color[3]

    @font_opacity.setter
    def font_opacity(self, val):
        self._font_color[3] = val * self.opacity

    @property
    def _draw_size(self):
        # if self.parent is None:
        #     return [0, 0]
        return blf.dimensions(self.font, self.text)

    def draw(self):
        super()._setup_draw()
        if self.parent.use_clipping:
            verts = self.parent._vertices
            blf.enable(self.font, blf.CLIPPING)
            blf.clipping(self.font, verts[0][0], verts[0][1], verts[2][0], verts[2][1])
        else:
            blf.disable(self.font, blf.CLIPPING)
        if self.wrap:
            blf.enable(self.font, blf.WORD_WRAP)
            blf.word_wrap(self.font, self.size[0])
        if self.shadow:
            col = self.shadow_color
            blf.enable(self.font, blf.SHADOW)
            blf.shadow(self.font, 0, col[0], col[1], col[2], col[3])
            blf.shadow_offset(self.font, int(self.shadow_offset[0]), int(self.shadow_offset[1]))
        blf.size(self.font, self.font_size)
        blf.color(self.font, self.font_color[0], self.font_color[1], self.font_color[2], self.font_color[3])
        charsize = blf.dimensions(self.font, 'A')
        lines = [t for t in self.text.split('\n')]
        if len(lines) > 1:
            for i, txt in enumerate(lines):
                pos = self._draw_pos.copy()
                dimensions = blf.dimensions(self.font, txt)
                lheight = (charsize[1] * self.line_height)
                if self.text_halign == 'center':
                    pos[0] -= (dimensions[0] * .5)
                elif self.text_halign == 'right':
                    pos[0] -= dimensions[0]
                if self.text_valign == 'top':
                    pos[1] -= lheight
                elif self.text_valign == 'center':
                    pos[1] += (.5 * lheight * (len(lines) - 1)) - (.5 * lheight)
                elif self.text_valign == 'bottom':
                    pos[1] += (lheight * (len(lines) -2))
                blf.position(self.font, pos[0], pos[1] - (charsize[1] * i * self.line_height), 0)
                blf.draw(self.font, txt)
        else:
            dimensions = blf.dimensions(self.font, self.text)
            pos = self._draw_pos.copy()
            if self.text_halign == 'center':
                pos[0] -= (dimensions[0] * .5)
            elif self.text_halign == 'right':
                pos[0] -= dimensions[0]
            if self.text_valign == 'top':
                pos[1] -= dimensions[1]
            elif self.text_valign == 'center':
                pos[1] -= (.5 * dimensions[1])
            blf.position(self.font, pos[0], pos[1], 0)
            blf.draw(self.font, self.text)

        blf.disable(self.font, blf.WORD_WRAP)
        blf.disable(self.font, blf.SHADOW)
        super().draw()



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.ui import LabelButton
from uplogic.utils import not_met
import bpy


class ULCreateUIButton(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.parent = None
        self.rel_pos = None
        self.pos = None
        self.rel_size = None
        self.size = None
        self.angle = None
        self.color = None
        self.border_width = None
        self.border_color = None
        self.text = None
        self.text_pos = None
        self.font = None
        self.font_size = None
        self.font_color = None
        self.line_height = None
        self.hover_color = None
        self._clicked = False
        self._hovering = False
        self._released = False
        self._widget = None
        self._done = False
        self.halign_type = 'left'
        self.valign_type = 'bottom'
        self.text_halign_type = 'left'
        self.text_valign_type = 'bottom'
        self.OUT = ULOutSocket(self, self._get_done)
        self.WIDGET = ULOutSocket(self, self._get_widget)
        self.CLICK = ULOutSocket(self, self._get_click)
        self.HOVER = ULOutSocket(self, self._get_hover)
        self.RELEASE = ULOutSocket(self, self._get_release)

    def _get_done(self):
        return self._done

    def _get_widget(self):
        return self._widget

    def _get_click(self):
        w = self._widget
        if w:
            return w._clicked
        return False

    def _get_hover(self):
        w = self._widget
        if w:
            return w._in_focus
        return False

    def _get_release(self):
        w = self._widget
        if w:
            return w._released
        return False

    def evaluate(self):
        self._done = False
        condition = self.get_input(self.condition)
        self._set_ready()
        if not_met(condition):
            return
        ipt = self.get_input
        parent = ipt(self.parent)
        rel_pos = ipt(self.rel_pos)
        pos = ipt(self.pos)
        rel_size = ipt(self.rel_size)
        size = ipt(self.size)
        color = ipt(self.color)
        border_width = ipt(self.border_width)
        border_color = ipt(self.border_color)
        hover_color = ipt(self.hover_color)
        text = ipt(self.text)
        text_pos = ipt(self.text_pos)
        font = ipt(self.font)
        font_size = ipt(self.font_size)
        font_color = ipt(self.font_color)
        line_height = ipt(self.line_height)

        font = bpy.data.fonts[font].filepath.replace('\\', '/') if font else 0

        self._widget = LabelButton(
            pos=pos,
            size=size,
            bg_color=color,
            relative={'pos': rel_pos, 'size': rel_size},
            hover_color=hover_color,
            text=text,
            text_pos=text_pos,
            font=font,
            font_size=font_size,
            font_color=font_color,
            line_height=line_height,
            halign_text=self.text_halign_type,
            valign_text=self.text_valign_type,
            border_width=border_width,
            border_color=border_color,
            halign=self.halign_type,
            valign=self.valign_type
        )
        if parent:
            parent.add_widget(self._widget)
        self._done = True
        
